LOGNAME=
NCORES= 4
RESTARTNAME= restart.txt
INSET=
BOX=
OUTSET=
OUTCUBUP=  10000000.0
PROGRESSLOG=

BMIN=
BMAJ=
BPA=
RMS=

NDISKS= 2

NUR= 4.00000
     RADI=0.0   10. 20. 30.
     VROT= 10.0
     Z0= 10.0
     SBR= 10.0
     INCL= 10.0
     PA= 10.0
     XPOS= 10.0
     YPOS= 10.0
     VSYS= 10.0
     SDIS= 10.0
     AZ1W= +1.80000E+02
     VROT_2= 10.0
     Z0_2= 10.0
     SBR_2= 10.0
     INCL_2= 10.0
     PA_2= 10.0
     XPOS_2= 10.0
     YPOS_2= 10.0
     VSYS_2= 10.0
     SDIS_2= 10.0
     AZ1P_2= +1.80000E+02
     AZ1W_2= +1.80000E+02
CONDISP= 10.0

LTYPE= 2
LTYPE_2= 2

CFLUX= 1e-5
CFLUX_2= 1e-5

PENALTY= 1
WEIGHT= 0
RADSEP= 0.1
INTY= 2
INDINTY = 0
ISEED= 8981
INIMODE= 3

FITMODE= 2
LOOPS= 6
MAXITER=
CALLITE=70
SIZE= 10


VARY=  XPOS 1:4 XPOS_2 1:4
VARINDX=
PARMAX=		       360
PARMIN=       	         0
MODERATE=     		 3
DELSTART=  	      1E-3
DELEND=	   	      1E-5
MINDELTA=  	      1E-5      	   

REGPARA=

TABLE=
DISTANCE=
REFRING= 5
BIGTABLE=
FRACTION=

TIRDEF= 1stfit.def
TIRSMO=
TIRACC= 5
TIRLEN=

COOLGAL=
COOLBEAM=

TILT=
BIGTILT=

INCLINO=
IN_REFINE= 5

GR_CONT=
RESTARTID = 0
