# -*- coding: future_fstrings -*-
import pyFAT_astro
#def initialize():
global H_0
H_0 = 69.7 #km/s/Mpc
global c
c=299792458                  #light speed in m/s
global pc
pc=3.086e+18                  #parsec in cm
global solar_mass
solar_mass=1.98855e30          #Solar mass in kg
global solar_luminosity
solar_luminosity = 3.828e26    # Bolometric Solar Luminosity in W
global HI_mass
HI_mass=1.6737236e-27             #Mass of hydrogen in kg
global G
G =  6.67430e-11     #m^3/kg*s^2
global Gsol
Gsol = G/(1000.**3)*solar_mass #km^3/M_sol*s^2
